@@ -158,11 +158,18 @@ cm_certread_n_main(int fd, struct cm_store_ca *ca, struct cm_store_entry *entry,
 		_exit(status);
 	}
     /* Re-open the database with modules enabled */
-	NSS_ShutdownContext(ctx);
+	if (NSS_ShutdownContext(ctx) != SECSuccess) {
+		cm_log(0, "Error shutting down NSS.\n");
+		_exit(1);
+	}
 	ctx = NSS_InitContext(entry->cm_cert_storage_location,
 			      NULL, NULL, NULL, NULL,
 			      (readwrite ? 0 : NSS_INIT_READONLY) |
 			      NSS_INIT_NOROOTINIT);
+	if (ctx == NULL) {
+		cm_log(0, "Unable to initialize NSS.\n");
+		_exit(1);
+	}
 	es = util_n_fips_hook();
 	if (es != NULL) {
 		cm_log(1, "Error putting NSS into FIPS mode: %s\n", es);
