@@ -175,6 +175,11 @@ cm_find_default_naming_context(LDAP *ld, char **basedn)
 	     lmsg = ldap_next_entry(ld, lmsg)) {
 		lbvalues = ldap_get_values_len(ld, lmsg,
 					       lncattrs[0]);
+		/* There should be only one defaultNamingContext so once we
+		 * have a value we're done. */
+		if (*basedn != NULL) {
+			break;
+		}
 		if (lbvalues == NULL) {
 			continue;
 		}
@@ -771,6 +776,7 @@ main(int argc, const char **argv)
 			printf(_("Unable to determine principal name for "
 				 "signing request.\n"));
 			poptPrintUsage(pctx, stdout, 0);
+			free(reqprinc);
 			return CM_SUBMIT_STATUS_UNCONFIGURED;
 		}
 		if ((profile == NULL) &&
@@ -813,6 +819,9 @@ main(int argc, const char **argv)
 				       CM_SUBMIT_CSR_ENV);
 			}
 			free(csr);
+			free(profile);
+			free(issuer);
+			free(reqprinc);
 			poptPrintUsage(pctx, stdout, 0);
 			return CM_SUBMIT_STATUS_UNCONFIGURED;
 		}
@@ -855,6 +864,10 @@ main(int argc, const char **argv)
 			fprintf(stderr,
 				"Error reading password from \"%s\": %s.\n",
 				pwdfile, strerror(errno));
+			free(csr);
+			free(profile);
+			free(issuer);
+			free(reqprinc);
 			return CM_SUBMIT_STATUS_UNCONFIGURED;
 		}
 	}
@@ -867,6 +880,10 @@ main(int argc, const char **argv)
 		if ((uid != NULL) || (pwd != NULL)) {
 			fprintf(stderr,
 				"Both -u and -W/-w options should be specified.\n");
+			free(csr);
+			free(profile);
+			free(issuer);
+			free(reqprinc);
 			return CM_SUBMIT_STATUS_UNCONFIGURED;
 		}
 	}
@@ -901,6 +918,10 @@ main(int argc, const char **argv)
 			}
 		}
 		free(kerr);
+		free(csr);
+		free(profile);
+		free(issuer);
+		free(reqprinc);
 		switch (kret) {
 		case KRB5_KDC_UNREACH:
 		case KRB5_REALM_CANT_RESOLVE:
@@ -920,10 +941,12 @@ main(int argc, const char **argv)
 				      basedn, uid, pwd, csr, reqprinc, profile,
 				      issuer);
 		free(csr);
+		free(profile);
+		free(issuer);
+		free(reqprinc);
 		return ret;
 	} else
 	if (strcasecmp(mode, CM_OP_FETCH_ROOTS) == 0) {
-		free(csr);
 		return fetch_roots(server, ldap_uri_cmd, ldap_uri, host,
 				   uid, pwd, domain, basedn);
 	}
