@@ -199,23 +199,22 @@ int
 main(int argc, const char **argv)
 {
 	const char *url = NULL, *results = NULL, *results2 = NULL;
-	struct cm_submit_h_context *hctx;
+	struct cm_submit_h_context *hctx = NULL;
 	int c, verbose = 0, results_length = 0, results_length2 = 0, i;
 	int prefer_non_renewal = 0, can_renewal = 0;
 	int response_code = 0, response_code2 = 0;
 	enum known_ops op = op_unset;
 	const char *id = NULL, *cainfo = NULL;
-	char *poptarg;
 	char *message = NULL, *rekey_message = NULL;
 	const char *mode = NULL, *content_type = NULL, *content_type2 = NULL;
 	void *ctx;
 	char *params = "", *params2 = NULL, *racert = NULL, *cacert = NULL;
-	char **othercerts = NULL, *cert1 = NULL, *cert2 = NULL, *certs = NULL;
+	char **certothers = NULL, *certleaf = NULL, *certtop = NULL, *certs = NULL;
 	char **racertp, **cacertp, *dracert = NULL, *dcacert = NULL;
 	char buf[LINE_MAX] = "";
 	const unsigned char **buffers = NULL;
 	size_t n_buffers = 0, *lengths = NULL, j;
-	const char *cacerts[3], **racerts;
+	const char *root[3], **othercerts;
 	dbus_bool_t missing_args = FALSE;
 	char *sent_tx, *tx, *msgtype, *pkistatus, *failinfo, *s, *tmp1, *tmp2;
 	unsigned char *sent_nonce, *sender_nonce, *recipient_nonce, *payload;
@@ -224,17 +223,19 @@ main(int argc, const char **argv)
 	size_t payload_length;
 	long error;
 	PKCS7 *p7;
-	poptContext pctx;
+	int rval = CM_SUBMIT_STATUS_UNCONFIGURED;
+	poptContext pctx = NULL;
 	struct poptOption popts[] = {
 		{"url", 'u', POPT_ARG_STRING, &url, 0, "service location", "URL"},
 		{"ca-identifier", 'i', POPT_ARG_STRING, &id, 0, "name to use when querying for capabilities", "IDENTIFIER"},
 		{"retrieve-ca-capabilities", 'c', POPT_ARG_NONE, NULL, 'c', "make a GetCACaps request", NULL},
-		{"retrieve-ca-certificates", 'C', POPT_ARG_NONE, NULL, 'C', "make GetCACert/GetCAChain requests", NULL},
+		{"retrieve-ca-certificates", 'C', POPT_ARG_NONE, NULL, 'C', "make GetCACert request", NULL},
 		{"get-initial-cert", 'g', POPT_ARG_NONE, NULL, 'g', "send a PKIOperation pkiMessage", NULL},
 		{"pki-message", 'p', POPT_ARG_NONE, NULL, 'p', "send a PKIOperation pkiMessage", NULL},
 		{"racert", 'r', POPT_ARG_STRING, NULL, 'r', "the RA certificate, used for encrypting requests", "FILENAME"},
-		{"cacert", 'R', POPT_ARG_STRING, NULL, 'R', "the CA certificate, used for verifying responses", "FILENAME"},
+		{"cacert", 'R', POPT_ARG_STRING, NULL, 'R', "the CA certificate, used for verifying TLS connections", "FILENAME"},
 		{"other-certs", 'I', POPT_ARG_STRING, NULL, 'I', "additional certificates", "FILENAME"},
+		{"signingca", 'N', POPT_ARG_STRING, NULL, 'N', "the CA certificate which signed the RA certificate", "FILENAME"},
 		{"non-renewal", 'n', POPT_ARG_NONE, &prefer_non_renewal, 0, "prefer to not use the SCEP Renewal feature", NULL},
 		{"verbose", 'v', POPT_ARG_NONE, NULL, 'v', NULL, NULL},
 		POPT_AUTOHELP
@@ -327,9 +328,10 @@ main(int argc, const char **argv)
 			racert = cm_submit_u_from_file(poptGetOptArg(pctx));
 			break;
 		case 'R':
-			poptarg = poptGetOptArg(pctx);
-			cainfo = strdup(poptarg);
-			cacert = cm_submit_u_from_file(poptarg);
+			cainfo = poptGetOptArg(pctx);
+			break;
+		case 'N':
+			cacert = cm_submit_u_from_file(poptGetOptArg(pctx));
 			break;
 		case 'I':
 			certs = cm_submit_u_from_file(poptGetOptArg(pctx));
@@ -338,7 +340,6 @@ main(int argc, const char **argv)
 	}
 	if (c != -1) {
 		poptPrintUsage(pctx, stdout, 0);
-		free(cainfo);
 		return CM_SUBMIT_STATUS_UNCONFIGURED;
 	}
 
@@ -368,11 +369,11 @@ main(int argc, const char **argv)
 		break;
 	case op_get_ca_caps:
 		/* Only step: read capabilities for the daemon. */
-		params = talloc_asprintf(ctx, "operation=" OP_GET_CA_CAPS);
+		params = talloc_asprintf(ctx, "operation=" OP_GET_CA_CAPS "&message=%s", id);
 		break;
 	case op_get_ca_certs:
 		/* First step: get the root certificate. */
-		params = talloc_asprintf(ctx, "operation=" OP_GET_CA_CERT);
+		params = talloc_asprintf(ctx, "operation=" OP_GET_CA_CERT "&message=%s", id);
 		break;
 	case op_get_cert_initial:
 		if ((racert == NULL) || (strlen(racert) == 0)) {
@@ -387,11 +388,11 @@ main(int argc, const char **argv)
 			}
 			if ((message == NULL) || (strlen(message) == 0)) {
 				printf(_("Error reading request.  Expected PKCS7 data containing a GetInitialCert pkiMessage, got nothing.\n"));
-				free(cainfo);
-				return CM_SUBMIT_STATUS_NEED_SCEP_MESSAGES;
+				rval = CM_SUBMIT_STATUS_NEED_SCEP_MESSAGES;
+				goto done;
 			}
 			/* First step: read capabilities for our use. */
-			params = talloc_asprintf(ctx, "operation=" OP_GET_CA_CAPS);
+			params = talloc_asprintf(ctx, "operation=" OP_GET_CA_CAPS "&message=%s", id);
 		}
 		break;
 	case op_pkcsreq:
@@ -407,11 +408,11 @@ main(int argc, const char **argv)
 			}
 			if ((message == NULL) || (strlen(message) == 0)) {
 				printf(_("Error reading request.  Expected PKCS7 data containing a PKCSReq pkiMessage, got nothing.\n"));
-				free(cainfo);
-				return CM_SUBMIT_STATUS_NEED_SCEP_MESSAGES;
+				rval = CM_SUBMIT_STATUS_NEED_SCEP_MESSAGES;
+				goto done;
 			}
 			/* First step: read capabilities for our use. */
-			params = talloc_asprintf(ctx, "operation=" OP_GET_CA_CAPS);
+			params = talloc_asprintf(ctx, "operation=" OP_GET_CA_CAPS "&message=%s", id);
 		}
 		break;
 	}
@@ -419,19 +420,23 @@ main(int argc, const char **argv)
 	/* Supply help output, if it's needed. */
 	if (missing_args) {
 		poptPrintUsage(pctx, stdout, 0);
-		free(cainfo);
-		return CM_SUBMIT_STATUS_UNCONFIGURED;
+		rval = CM_SUBMIT_STATUS_UNCONFIGURED;
+		goto done;
 	}
 
 	/* Check the rekey PKCSReq message, if we have one. */
 	if ((rekey_message != NULL) && (strlen(rekey_message) != 0)) {
 		tmp1 = cm_submit_u_base64_from_text(rekey_message);
 		tmp2 = cm_store_base64_as_bin(ctx, tmp1, -1, &c);
-		cm_pkcs7_verify_signed((unsigned char *) tmp2, c,
+		i = cm_pkcs7_verify_signed((unsigned char *) tmp2, c,
 				       NULL, NULL, NID_pkcs7_data, ctx, NULL,
 				       NULL, &msgtype, NULL, NULL,
 				       NULL, NULL,
 				       NULL, NULL, NULL, NULL);
+		if (i != 0) {
+			log_pkcs7_errors(0, "Error: failed to verify signature on "
+					"rekey PKCSReq.\n");
+		}
 		if ((msgtype == NULL) ||
 		    ((strcmp(msgtype, SCEP_MSGTYPE_PKCSREQ) != 0) &&
 		     (strcmp(msgtype, SCEP_MSGTYPE_GETCERTINITIAL) != 0))) {
@@ -453,11 +458,15 @@ main(int argc, const char **argv)
 	if ((message != NULL) && (strlen(message) != 0)) {
 		tmp1 = cm_submit_u_base64_from_text(message);
 		tmp2 = cm_store_base64_as_bin(ctx, tmp1, -1, &c);
-		cm_pkcs7_verify_signed((unsigned char *) tmp2, c,
+		i = cm_pkcs7_verify_signed((unsigned char *) tmp2, c,
 				       NULL, NULL, NID_pkcs7_data, ctx, NULL,
 				       &sent_tx, &msgtype, NULL, NULL,
 				       &sent_nonce, &sent_nonce_length,
 				       NULL, NULL, NULL, NULL);
+		if (i != 0) {
+			log_pkcs7_errors(0, "Error: failed to verify signature on "
+					"message.\n");
+		}
 		if ((msgtype == NULL) ||
 		    ((strcmp(msgtype, SCEP_MSGTYPE_PKCSREQ) != 0) &&
 		     (strcmp(msgtype, SCEP_MSGTYPE_GETCERTINITIAL) != 0))) {
@@ -496,7 +505,6 @@ main(int argc, const char **argv)
 				verbose > 1 ?
 				cm_submit_h_curl_verbose_on :
 				cm_submit_h_curl_verbose_off);
-	free(cainfo);
 	cm_submit_h_run(hctx);
 	content_type = cm_submit_h_result_type(hctx);
 	if (content_type == NULL) {
@@ -542,7 +550,8 @@ main(int argc, const char **argv)
 		}
 		if ((tmp2 == NULL) || (strlen(tmp2) == 0)) {
 			printf(_("Error reading request.  Expected PKCS7 data containing a GetInitialCert pkiMessage, got nothing.\n"));
-			return CM_SUBMIT_STATUS_NEED_SCEP_MESSAGES;
+			rval = CM_SUBMIT_STATUS_NEED_SCEP_MESSAGES;
+			goto done;
 		} else
 		if (verbose > 0) {
 			if (tmp2 == rekey_message) {
@@ -567,7 +576,8 @@ main(int argc, const char **argv)
 		}
 		if ((tmp2 == NULL) || (strlen(tmp2) == 0)) {
 			printf(_("Error reading request.  Expected PKCS7 data containing a PKCSReq pkiMessage, got nothing.\n"));
-			return CM_SUBMIT_STATUS_NEED_SCEP_MESSAGES;
+			rval = CM_SUBMIT_STATUS_NEED_SCEP_MESSAGES;
+			goto done;
 		} else
 		if (verbose > 0) {
 			if (tmp2 == rekey_message) {
@@ -629,7 +639,8 @@ main(int argc, const char **argv)
 			       cm_submit_h_result_code(hctx),
 			       url);
 		}
-		return CM_SUBMIT_STATUS_UNREACHABLE;
+		rval = CM_SUBMIT_STATUS_UNREACHABLE;
+		goto done;
 	}
 	switch (op) {
 	case op_unset:
@@ -642,16 +653,19 @@ main(int argc, const char **argv)
 			       response_code, url);
 			if (response_code == 500) {
 				/* The server might recover, right? */
-				return CM_SUBMIT_STATUS_UNREACHABLE;
+				rval = CM_SUBMIT_STATUS_UNREACHABLE;
+				goto done;
 			} else {
 				/* Maybe not? */
-				return CM_SUBMIT_STATUS_REJECTED;
+				rval = CM_SUBMIT_STATUS_REJECTED;
+				goto done;
 			}
 		}
 		if (results == NULL) {
 			printf(_("Internal error: no response to \"%s?%s\".\n"),
 			       url, params);
-			return CM_SUBMIT_STATUS_REJECTED;
+			rval = CM_SUBMIT_STATUS_REJECTED;
+			goto done;
 		}
 		break;
 	case op_get_cert_initial:
@@ -676,10 +690,12 @@ main(int argc, const char **argv)
 				fprintf(stderr, "Result is surprisingly large, "
 					"suppressing it.\n");
 			}
-			return CM_SUBMIT_STATUS_REJECTED;
+			rval = CM_SUBMIT_STATUS_REJECTED;
+			goto done;
 		}
 		printf("%s\n", results);
-		return CM_SUBMIT_STATUS_ISSUED;
+		rval = CM_SUBMIT_STATUS_ISSUED;
+		goto done;
 		break;
 	case op_get_ca_certs:
 		if ((strcasecmp(content_type,
@@ -688,7 +704,8 @@ main(int argc, const char **argv)
 				"application/x-x509-ca-ra-cert") != 0)) {
 			printf(_("Server reply was of unexpected MIME type "
 				 "\"%s\".\n"), content_type);
-			return CM_SUBMIT_STATUS_UNREACHABLE;
+			rval = CM_SUBMIT_STATUS_UNREACHABLE;
+			goto done;
 		}
 		if (racert == NULL) {
 			racertp = &racert;
@@ -701,7 +718,8 @@ main(int argc, const char **argv)
 						 n_buffers + 1);
 			if ((buffers == NULL) || (lengths == NULL)) {
 				fprintf(stderr, "Out of memory.\n");
-				return CM_SUBMIT_STATUS_UNREACHABLE;
+				rval = CM_SUBMIT_STATUS_UNREACHABLE;
+				goto done;
 			}
 			buffers[n_buffers] = (unsigned char *) racert;
 			lengths[n_buffers] = strlen(racert);
@@ -718,7 +736,8 @@ main(int argc, const char **argv)
 						 n_buffers + 1);
 			if ((buffers == NULL) || (lengths == NULL)) {
 				fprintf(stderr, "Out of memory.\n");
-				return CM_SUBMIT_STATUS_UNREACHABLE;
+				rval = CM_SUBMIT_STATUS_UNREACHABLE;
+				goto done;
 			}
 			buffers[n_buffers] = (unsigned char *) cacert;
 			lengths[n_buffers] = strlen(cacert);
@@ -732,7 +751,8 @@ main(int argc, const char **argv)
 						 n_buffers + 1);
 			if ((buffers == NULL) || (lengths == NULL)) {
 				fprintf(stderr, "Out of memory.\n");
-				return CM_SUBMIT_STATUS_UNREACHABLE;
+				rval = CM_SUBMIT_STATUS_UNREACHABLE;
+				goto done;
 			}
 			buffers[n_buffers] = (unsigned char *) results;
 			lengths[n_buffers] = results_length;
@@ -746,7 +766,8 @@ main(int argc, const char **argv)
 						 n_buffers + 1);
 			if ((buffers == NULL) || (lengths == NULL)) {
 				fprintf(stderr, "Out of memory.\n");
-				return CM_SUBMIT_STATUS_UNREACHABLE;
+				rval = CM_SUBMIT_STATUS_UNREACHABLE;
+				goto done;
 			}
 			buffers[n_buffers] = (unsigned char *) results2;
 			lengths[n_buffers] = results_length2;
@@ -841,90 +862,127 @@ main(int argc, const char **argv)
 						 n_buffers + 1);
 			if ((buffers == NULL) || (lengths == NULL)) {
 				fprintf(stderr, "Out of memory.\n");
-				return CM_SUBMIT_STATUS_UNREACHABLE;
+				rval = CM_SUBMIT_STATUS_UNREACHABLE;
+				goto done;
 			}
 			buffers[n_buffers] = (unsigned char *) results2;
 			lengths[n_buffers] = results_length2;
 			n_buffers++;
 		}
 		if (cm_pkcs7_parsev(CM_PKCS7_LEAF_PREFER_ENCRYPT, ctx,
-				    racertp, cacertp, &othercerts,
+				    racertp, cacertp, &certothers,
 				    NULL, NULL,
 				    n_buffers, buffers, lengths) == 0) {
 			if (racert != NULL) {
 				printf("%s", racert);
 				if (cacert != NULL) {
 					printf("%s", cacert);
-					if (othercerts != NULL) {
+					if (certothers != NULL) {
 						for (c = 0;
-						     othercerts[c] != NULL;
+						     certothers[c] != NULL;
 						     c++) {
 							printf("%s",
-							       othercerts[c]);
+							       certothers[c]);
 						}
 					}
 					if ((dracert != NULL) &&
-					    (cert_among(dracert, racert, cacert, othercerts) != 0)) {
+					    (cert_among(dracert, racert, cacert, certothers) != 0)) {
 						printf("%s", dracert);
 					}
 					if ((dcacert != NULL) &&
-					    (cert_among(dcacert, racert, cacert, othercerts) != 0)) {
+					    (cert_among(dcacert, racert, cacert, certothers) != 0)) {
 						printf("%s", dcacert);
 					}
 				}
 			}
-			talloc_free(ctx);
-			return CM_SUBMIT_STATUS_ISSUED;
+			rval = CM_SUBMIT_STATUS_ISSUED;
+			goto done;
 		} else {
-			talloc_free(ctx);
-			return CM_SUBMIT_STATUS_UNREACHABLE;
+			rval = CM_SUBMIT_STATUS_UNREACHABLE;
+			goto done;
 		}
 		break;
 	case op_get_cert_initial:
 	case op_pkcsreq:
 		if ((content_type2 != NULL) && (strcasecmp(content_type2,
 			       "application/x-pki-message") == 0)) {
-			memset(&cacerts, 0, sizeof(cacerts));
-			cacerts[0] = cacert ? cacert : racert;
-			cacerts[1] = cacert ? racert : NULL;
-			cacerts[2] = NULL;
-			racerts = NULL;
+			/*
+			 * At this point, we have:
+			 * - zero or more ra certs; and
+			 * - zero or more ca certificates; and
+			 * - zero or more other certificates; that
+			 * need to be reordered so that the leaf
+			 * certificates go first, the ca certificates
+			 * are separated into a seperate certificate
+			 * store, and the other certificates go after
+			 * the leaf certificates.
+			 *
+			 * To do this we put cacert into the ca store,
+			 * the racert at the top of the othercerts list.
+			 * Then we parse certs, placing all ca certs
+			 * we find into the ca store, and all other
+			 * certs we find after the racert.
+			 *
+			 * As a limitation of cm_pkcs7_parse(), we
+			 * can only isolate one ca certificate in the
+			 * list of other certificates.
+			 */
+			/* handle the other certs */
 			if ((certs != NULL) &&
 			    (cm_pkcs7_parse(0, ctx,
-					    &cert1, &cert2, &othercerts,
+					    &certleaf, &certtop, &certothers,
 					    NULL, NULL,
 					    (const unsigned char *) certs,
 					    strlen(certs), NULL) == 0)) {
+				/* Special case for IPA which uses dogtag which signs SCEP
+				 * certs using the CA cert and the typical way to get
+				 * verification to work is to use -I /etc/ipa/ca.crt.
+				 * Because cm_pkcs7_parse explicitly doesn't allow
+				 * certleaf to equal certtop we end up with no CAs so verification
+				 * fails.
+				 * 
+				 * So if cacert and certleaf are both NULL and certtop is
+				 * self-signed then assume the IPA case and set certtop equal
+				 * to certleaf.
+				 */
+				if ((cacert == NULL) && (certtop == NULL) && (certleaf != NULL)) {
+					if (cm_selfsigned(certleaf) == 0) {
+						certtop = certleaf;
+					}
+				}
+				memset(&root, 0, sizeof(root));
+				root[0] = cacert ? cacert : certtop ? certtop : NULL;
+				root[1] = cacert ? certtop : NULL;
+				root[2] = NULL;
 				for (c = 0;
-				     (othercerts != NULL) &&
-				     (othercerts[c] != NULL);
+				     (certothers != NULL) &&
+				     (certothers[c] != NULL);
 				     c++) {
 					continue;
 				}
-				racerts = talloc_array_ptrtype(ctx, racerts, c + 5);
-				for (c = 0;
-				     (othercerts != NULL) &&
-				     (othercerts[c] != NULL);
-				     c++) {
-					racerts[c] = othercerts[c];
-				}
-				if (cacert != NULL) {
-					racerts[c++] = cacert;
-				}
-				if (cert1 != NULL) {
-					racerts[c++] = cert1;
-				}
-				if (cert2 != NULL) {
-					racerts[c++] = cert2;
-				}
+				othercerts = talloc_array_ptrtype(ctx, othercerts, c + 3);
+				c = 0;
 				if (racert != NULL) {
-					racerts[c++] = racert;
+					othercerts[c++] = racert;
 				}
-				racerts[c++] = NULL;
+				if (certleaf != NULL) {
+					othercerts[c++] = certleaf;
+				}
+				while (certothers != NULL && *certothers != NULL) {
+					othercerts[c++] = *certothers++;
+				}
+				othercerts[c++] = NULL;
+			}
+			else {
+				root[0] = cacert;
+				root[1] = NULL;
+				othercerts = talloc_array_ptrtype(ctx, othercerts, 2);
+				othercerts[0] = racert ? racert : NULL;
+				othercerts[1] = NULL;
 			}
 			ERR_clear_error();
 			i = cm_pkcs7_verify_signed((unsigned char *) results2, results_length2,
-						   cacerts, racerts,
+						   root, othercerts,
 						   NID_pkcs7_data, ctx, NULL,
 						   &tx, &msgtype, &pkistatus, &failinfo,
 						   &sender_nonce, &sender_nonce_length,
@@ -932,56 +990,66 @@ main(int argc, const char **argv)
 						   &payload, &payload_length);
 			if (i != 0) {
 				printf(_("Error: failed to verify signature on "
-					 "server response.\n"));
-				cm_log(1, "Error: failed to verify signature on "
-					 "server response.\n");
-				while ((error = ERR_get_error()) != 0) {
+						 "server response. "));
+				error = ERR_peek_last_error();
+				if (error != 0) {
 					memset(buf, '\0', sizeof(buf));
 					ERR_error_string_n(error, buf, sizeof(buf));
-					cm_log(1, "%s\n", buf);
+					printf("%s", buf);
 				}
+				printf("\n");
+				log_pkcs7_errors(0, "Error: failed to verify signature on "
+						  "server response.\n");
 				s = cm_store_base64_from_bin(ctx, (unsigned char *) results2,
 							     results_length2);
 				s = cm_submit_u_pem_from_base64("PKCS7", 0, s);
 				fprintf(stderr, "%s", s);
 				cm_log(1, "%s", s);
 				free(s);
-				return CM_SUBMIT_STATUS_UNREACHABLE;
+				rval = CM_SUBMIT_STATUS_UNREACHABLE;
+				goto done;
 			}
 			if ((msgtype == NULL) ||
 			    (strcmp(msgtype, SCEP_MSGTYPE_CERTREP) != 0)) {
 				printf(_("Error: reply was not a CertRep (%s).\n"),
 				       msgtype ? msgtype : "none");
-				return CM_SUBMIT_STATUS_UNREACHABLE;
+				rval = CM_SUBMIT_STATUS_UNREACHABLE;
+				goto done;
 			}
 			if (tx == NULL) {
 				printf(_("Error: reply is missing transactionId.\n"));
-				return CM_SUBMIT_STATUS_UNREACHABLE;
+				rval = CM_SUBMIT_STATUS_UNREACHABLE;
+				goto done;
 			}
 			if (sent_tx != NULL) {
 				if (strcmp(sent_tx, tx) != 0) {
 					printf(_("Error: reply contains a "
 						 "different transactionId.\n"));
-					return CM_SUBMIT_STATUS_UNREACHABLE;
+					rval = CM_SUBMIT_STATUS_UNREACHABLE;
+					goto done;
 				}
 			}
 			if (pkistatus == NULL) {
 				printf(_("Error: reply is missing pkiStatus.\n"));
-				return CM_SUBMIT_STATUS_UNREACHABLE;
+				rval = CM_SUBMIT_STATUS_UNREACHABLE;
+				goto done;
 			}
 			if (recipient_nonce == NULL) {
 				printf(_("Error: reply is missing recipientNonce.\n"));
-				return CM_SUBMIT_STATUS_UNREACHABLE;
+				rval = CM_SUBMIT_STATUS_UNREACHABLE;
+				goto done;
 			}
-			if ((recipient_nonce_length != sent_nonce_length) ||
+			if (sent_nonce && ((recipient_nonce_length != sent_nonce_length) ||
 			    (memcmp(recipient_nonce, sent_nonce,
-				    sent_nonce_length) != 0)) {
+				    sent_nonce_length) != 0))) {
 				printf(_("Error: reply nonce doesn't match request.\n"));
-				return CM_SUBMIT_STATUS_UNREACHABLE;
+				rval = CM_SUBMIT_STATUS_UNREACHABLE;
+				goto done;
 			}
 			if (sender_nonce == NULL) {
 				printf(_("Error: reply is missing senderNonce.\n"));
-				return CM_SUBMIT_STATUS_UNREACHABLE;
+				rval = CM_SUBMIT_STATUS_UNREACHABLE;
+				goto done;
 			}
 			if (strcmp(pkistatus, SCEP_PKISTATUS_PENDING) == 0) {
 				if (verbose > 0) {
@@ -991,7 +1059,8 @@ main(int argc, const char **argv)
 				s = cm_store_base64_from_bin(ctx, sender_nonce,
 							     sender_nonce_length);
 				printf("%s\n", s);
-				return CM_SUBMIT_STATUS_WAIT;
+				rval = CM_SUBMIT_STATUS_WAIT;
+				goto done;
 			} else
 			if (strcmp(pkistatus, SCEP_PKISTATUS_FAILURE) == 0) {
 				if (verbose > 0) {
@@ -1039,7 +1108,8 @@ main(int argc, const char **argv)
 					printf(_("Server returned failure code \"%s\".\n"),
 					       failinfo);
 				}
-				return CM_SUBMIT_STATUS_REJECTED;
+				rval = CM_SUBMIT_STATUS_REJECTED;
+				goto done;
 			} else
 			if (strcmp(pkistatus, SCEP_PKISTATUS_SUCCESS) == 0) {
 				if (verbose > 0) {
@@ -1049,88 +1119,77 @@ main(int argc, const char **argv)
 				p7 = d2i_PKCS7(NULL, &u, payload_length);
 				if (p7 == NULL) {
 					printf(_("Error: couldn't parse signed-data.\n"));
-					while ((error = ERR_get_error()) != 0) {
-						memset(buf, '\0', sizeof(buf));
-						ERR_error_string_n(error, buf, sizeof(buf));
-						cm_log(1, "%s\n", buf);
-					}
+					log_pkcs7_errors(0, "Error: couldn't parse signed-data.\n");
 					s = cm_store_base64_from_bin(ctx,
 								     (unsigned char *) results2,
 								     results_length2);
 					s = cm_submit_u_pem_from_base64("PKCS7", 0, s);
 					fprintf(stderr, "Full reply:\n%s", s);
 					free(s);
-					return CM_SUBMIT_STATUS_UNREACHABLE;
+					rval = CM_SUBMIT_STATUS_UNREACHABLE;
+					goto done;
 				}
 				if (!PKCS7_type_is_enveloped(p7)) {
 					printf(_("Error: signed-data payload is not enveloped-data.\n"));
-					while ((error = ERR_get_error()) != 0) {
-						memset(buf, '\0', sizeof(buf));
-						ERR_error_string_n(error, buf, sizeof(buf));
-						cm_log(1, "%s\n", buf);
-					}
+					log_pkcs7_errors(0, "Error: signed-data payload is not "
+								"enveloped-data.\n");
 					s = cm_store_base64_from_bin(ctx,
 								     (unsigned char *) results2,
 								     results_length2);
 					s = cm_submit_u_pem_from_base64("PKCS7", 0, s);
 					fprintf(stderr, "Full reply:\n%s", s);
 					free(s);
-					return CM_SUBMIT_STATUS_UNREACHABLE;
-				}
-				if (!PKCS7_type_is_enveloped(p7)) {
-					printf(_("Error: signed-data payload is not enveloped-data.\n"));
-					while ((error = ERR_get_error()) != 0) {
-						memset(buf, '\0', sizeof(buf));
-						ERR_error_string_n(error, buf, sizeof(buf));
-						cm_log(1, "%s\n", buf);
-					}
-					s = cm_store_base64_from_bin(ctx,
-								     (unsigned char *) results2,
-								     results_length2);
-					s = cm_submit_u_pem_from_base64("PKCS7", 0, s);
-					fprintf(stderr, "Full reply:\n%s", s);
-					free(s);
-					return CM_SUBMIT_STATUS_UNREACHABLE;
+					rval = CM_SUBMIT_STATUS_UNREACHABLE;
+					goto done;
 				}
 				if ((p7->d.enveloped == NULL) ||
 				    (p7->d.enveloped->enc_data == NULL) ||
 				    (p7->d.enveloped->enc_data->content_type == NULL) ||
 				    (OBJ_obj2nid(p7->d.enveloped->enc_data->content_type) != NID_pkcs7_data)) {
 					printf(_("Error: enveloped-data payload is not data.\n"));
-					while ((error = ERR_get_error()) != 0) {
-						memset(buf, '\0', sizeof(buf));
-						ERR_error_string_n(error, buf, sizeof(buf));
-						cm_log(1, "%s\n", buf);
-					}
+					log_pkcs7_errors(0, "Error: enveloped-data payload is "
+								"not data.\n");
 					s = cm_store_base64_from_bin(ctx,
 								     (unsigned char *) results2,
 								     results_length2);
 					s = cm_submit_u_pem_from_base64("PKCS7", 0, s);
 					fprintf(stderr, "Full reply:\n%s", s);
 					free(s);
-					return CM_SUBMIT_STATUS_UNREACHABLE;
+					rval = CM_SUBMIT_STATUS_UNREACHABLE;
+					goto done;
 				}
 				s = cm_store_base64_from_bin(ctx, payload,
 							     payload_length);
 				s = cm_submit_u_pem_from_base64("PKCS7", 0, s);
 				printf("%s", s);
 				free(s);
-				return CM_SUBMIT_STATUS_ISSUED;
+				rval = CM_SUBMIT_STATUS_ISSUED;
+				goto done;
 			} else {
 				if (verbose > 0) {
 					fprintf(stderr, "SCEP status is \"%s\".\n", pkistatus);
 				}
 				printf(_("Error: pkiStatus \"%s\" not recognized.\n"),
 				       pkistatus);
-				return CM_SUBMIT_STATUS_UNREACHABLE;
+				rval = CM_SUBMIT_STATUS_UNREACHABLE;
+				goto done;
 			}
 		} else {
 			printf(_("Server reply was of unexpected MIME type "
 				 "\"%s\".\n"), content_type);
 			printf("Full reply:\n%.*s", results_length2, results2);
-			return CM_SUBMIT_STATUS_UNREACHABLE;
+			rval = CM_SUBMIT_STATUS_UNREACHABLE;
+			goto done;
 		}
 		break;
 	}
-	return CM_SUBMIT_STATUS_UNCONFIGURED;
+
+done:
+	if (pctx) {
+		poptFreeContext(pctx);
+	}
+	free(id);
+	cm_submit_h_cleanup(hctx);
+	talloc_free(ctx);
+	return rval;
 }
