@@ -17,22 +17,36 @@
 
 #include "config.h"
 
+#include <sys/types.h>
 #include <errno.h>
 #include <stdlib.h>
 #include <string.h>
+#include <unistd.h>
+#include <signal.h>
 
 #include <talloc.h>
 #include <tevent.h>
 
+#include <krb5.h>
+
 #include <dbus/dbus.h>
 
+#ifdef HAVE_OPENSSL
+#include <openssl/rand.h>
+#endif
+#ifdef HAVE_GMP
+#include <gmp.h>
+#endif
+
 #include "cm.h"
 #include "log.h"
+#include "submit-u.h"
 #include "tdbus.h"
 #include "tdbush.h"
 #include "tdbusm.h"
 
 struct tdbus_connection {
+	DBusServer *server;
 	DBusConnection *conn;
 	enum cm_tdbus_type conn_type;
 	struct tdbus_watch {
@@ -58,7 +72,10 @@ struct tdbus_connection {
 	void *data;
 };
 
-static int cm_tdbus_setup_connection(struct tdbus_connection *tdb, DBusError *);
+static int cm_tdbus_setup_public_connection(struct tdbus_connection *tdb,
+					    DBusConnection *conn,
+					    const char *bus_desc,
+					    DBusError *error);
 
 static void
 cm_tdbus_dispatch_status(DBusConnection *conn, DBusDispatchStatus new_status,
@@ -124,6 +141,8 @@ cm_tdbus_queue_fd(struct tevent_context *ec, struct tdbus_watch *watch,
 {
 	struct tdbus_dwatch *dwatch;
 	int newtflags, dflags;
+	char flags[20] = "";
+
 	newtflags = 0;
 	dwatch = watch->dwatches;
 	while (dwatch != NULL) {
@@ -135,12 +154,23 @@ cm_tdbus_queue_fd(struct tevent_context *ec, struct tdbus_watch *watch,
 		dwatch = dwatch->next;
 	}
 	if (newtflags != 0) {
-		cm_log(5, "Queuing FD %d for 0x%02x.\n", watch->fd, newtflags);
+		if (newtflags & TEVENT_FD_READ) {
+			strcpy(flags, "Read");
+		}
+		if (newtflags & TEVENT_FD_WRITE) {
+			if (strlen(flags) > 0) {
+				strcat(flags, "-");
+			}
+			strcat(flags, "Write");
+		}
 		watch->tfd = tevent_add_fd(ec, watch, watch->fd, newtflags,
 					   handler, watch);
+		cm_log(5, "Queuing FD %d for %s for %p:%p.\n", watch->fd,
+		       flags, watch->conn, watch->tfd);
 	} else {
-		cm_log(5, "Not queuing FD %d.\n", watch->fd);
 		watch->tfd = NULL;
+		cm_log(5, "Not queuing FD %d for %p.\n", watch->fd,
+		       watch->conn);
 	}
 }
 
@@ -151,22 +181,49 @@ cm_tdbus_handle_fd(struct tevent_context *ec, struct tevent_fd *tfd,
 	struct tdbus_watch *watch;
 	struct tdbus_dwatch *dwatch;
 	int dflags;
+	char flags[20] = "";
+
 	watch = pvt;
+	dflags = cm_tdbus_watch_flags_for_tfd_flags(tflags);
+	if (tflags & TEVENT_FD_READ) {
+		strcpy(flags, "Read");
+	}
+	if (tflags & TEVENT_FD_WRITE) {
+		if (strlen(flags) > 0) {
+			strcat(flags, "-");
+		}
+		strcat(flags, "Write");
+	}
+	cm_log(5, "Dequeuing FD %d for %s for %p:%p.\n",
+	       watch->fd, flags, watch->conn, watch->tfd);
 	talloc_free(watch->tfd);
 	watch->tfd = NULL;
 	dwatch = watch->dwatches;
-	dflags = cm_tdbus_watch_flags_for_tfd_flags(tflags);
 	while (dwatch != NULL) {
 		if (dwatch->active) {
-			cm_log(5, "Handling D-Bus traffic on %d.\n", watch->fd);
+			cm_log(5, "Handling D-Bus traffic (%s) on FD %d for "
+			       "%p.\n", flags, watch->fd, watch->conn);
 			if ((dflags & dwatch->dflags) != 0) {
 				dbus_watch_handle(dwatch->watch,
 						  dflags & dwatch->dflags);
 				break;
 			}
+		} else {
+			cm_log(5, "Skipping disabled %d handler on FD %d for "
+			       "%p.\n", dwatch->dflags, watch->fd, watch->conn);
 		}
 		dwatch = dwatch->next;
 	}
+	if (dwatch == NULL) {
+		cm_log(5, "Unexpected D-Bus traffic (%s) on FD %d for %p:%p.\n",
+		       flags, watch->fd, watch->conn, tfd);
+	}
+	if (watch->tfd != NULL) {
+		cm_log(5, "Dequeuing FD %d for %s for %p:%p.\n",
+		       watch->fd, flags, watch->conn, watch->tfd);
+		talloc_free(watch->tfd);
+		watch->tfd = NULL;
+	}
 	cm_tdbus_queue_fd(ec, watch, cm_tdbus_handle_fd);
 }
 
@@ -198,9 +255,22 @@ cm_tdbus_watch_add(DBusWatch *watch, void *data)
 	struct tdbus_watch *tdb_watch;
 	struct tdbus_dwatch *tdb_dwatch;
 	int fd;
+	char flags[20] = "";
+
 	conn = data;
 	fd = cm_tdbus_watch_get_fd(watch);
-	cm_log(5, "Adding DBus watch on %d.\n", fd);
+	if (dbus_watch_get_flags(watch) & DBUS_WATCH_READABLE) {
+		strcpy(flags, "Read");
+	}
+	if (dbus_watch_get_flags(watch) & DBUS_WATCH_WRITABLE) {
+		if (strlen(flags) > 0) {
+			strcat(flags, "-");
+		}
+		strcat(flags, "Write");
+	}
+	cm_log(5, "Adding %sabled DBus watch on FD %d (for %s) for %p.\n",
+	       dbus_watch_get_enabled(watch) ? "en" : "dis", fd, flags,
+	       data);
 	/* Find the tevent watch for this fd. */
 	tdb_watch = conn->watches;
 	while (tdb_watch != NULL) {
@@ -211,7 +281,8 @@ cm_tdbus_watch_add(DBusWatch *watch, void *data)
 	}
 	/* If we couldn't find one, add it. */
 	if (tdb_watch == NULL) {
-		cm_log(5, "Adding a new tevent FD for %d.\n", fd);
+		cm_log(5, "Adding a watch group for FD %d for %p.\n", fd,
+		       data);
 		tdb_watch = talloc_ptrtype(conn, tdb_watch);
 		if (tdb_watch == NULL) {
 			return FALSE;
@@ -236,7 +307,10 @@ cm_tdbus_watch_add(DBusWatch *watch, void *data)
 	tdb_dwatch->next = tdb_watch->dwatches;
 	tdb_watch->dwatches = tdb_dwatch;
 	/* (Re-)queue the tfd. */
+	cm_log(5, "Dequeuing FD %d for %p:%p.\n",
+	       tdb_watch->fd, tdb_watch->conn, tdb_watch->tfd);
 	talloc_free(tdb_watch->tfd);
+	tdb_watch->tfd = NULL;
 	cm_tdbus_queue_fd(talloc_parent(conn), tdb_watch, cm_tdbus_handle_fd);
 	return TRUE;
 }
@@ -250,7 +324,8 @@ cm_tdbus_watch_remove(DBusWatch *watch, void *data)
 	int fd;
 	conn = data;
 	fd = cm_tdbus_watch_get_fd(watch);
-	cm_log(5, "Removing a DBus watch for %d.\n", fd);
+	cm_log(5, "Removing a DBus watch for FD %d (for %u) for %p.\n", fd,
+	       dbus_watch_get_flags(watch), data);
 	/* Find the tevent watch for this fd. */
 	tdb_watch = conn->watches;
 	while (tdb_watch != NULL) {
@@ -260,6 +335,7 @@ cm_tdbus_watch_remove(DBusWatch *watch, void *data)
 		tdb_watch = tdb_watch->next;
 	}
 	if (tdb_watch == NULL) {
+		cm_log(5, "No matching watch found.\n");
 		return;
 	}
 	/* Find the watch in the list of dwatches. */
@@ -281,7 +357,10 @@ cm_tdbus_watch_remove(DBusWatch *watch, void *data)
 		prev = tdb_dwatch;
 	}
 	/* (Re-)queue the tfd. */
+	cm_log(5, "Dequeuing FD %d for %p:%p.\n",
+	       tdb_watch->fd, tdb_watch->conn, tdb_watch->tfd);
 	talloc_free(tdb_watch->tfd);
+	tdb_watch->tfd = NULL;
 	cm_tdbus_queue_fd(talloc_parent(conn), tdb_watch, cm_tdbus_handle_fd);
 }
 
@@ -294,6 +373,8 @@ cm_tdbus_watch_toggle(DBusWatch *watch, void *data)
 	int fd;
 	conn = data;
 	fd = cm_tdbus_watch_get_fd(watch);
+	cm_log(5, "Toggling a DBus watch for FD %d (for %u) for "
+	       "%p.\n", fd, dbus_watch_get_flags(watch), conn);
 	/* Find the tevent watch for this fd. */
 	tdb_watch = conn->watches;
 	while (tdb_watch != NULL) {
@@ -303,6 +384,7 @@ cm_tdbus_watch_toggle(DBusWatch *watch, void *data)
 		tdb_watch = tdb_watch->next;
 	}
 	if (tdb_watch == NULL) {
+		cm_log(5, "No matching watch found.\n");
 		return;
 	}
 	/* Find the watch in the list of dwatches. */
@@ -310,12 +392,17 @@ cm_tdbus_watch_toggle(DBusWatch *watch, void *data)
 	while (tdb_dwatch != NULL) {
 		if (tdb_dwatch->watch == watch) {
 			tdb_dwatch->active = dbus_watch_get_enabled(watch);
+			cm_log(5, "Watch %sabled.\n",
+			       tdb_dwatch->active ?  "en" : "dis");
 			break;
 		}
 		tdb_dwatch = tdb_dwatch->next;
 	}
 	/* (Re-)queue the tfd. */
+	cm_log(5, "Dequeuing FD %d for %p:%p.\n",
+	       tdb_watch->fd, tdb_watch->conn, tdb_watch->tfd);
 	talloc_free(tdb_watch->tfd);
+	tdb_watch->tfd = NULL;
 	cm_tdbus_queue_fd(talloc_parent(conn), tdb_watch, cm_tdbus_handle_fd);
 }
 
@@ -436,55 +523,24 @@ cm_tdbus_timeout_cleanup(void *data)
 }
 
 static void
-cm_tdbus_reconnect(struct tevent_context *ec, struct tevent_timer *timer,
+cm_tdbus_disconnected(struct tevent_context *ec, struct tevent_timer *timer,
 		   struct timeval current_time, void *pvt)
 {
-	const char *bus_desc;
 	struct tdbus_connection *tdb;
-	struct timeval later;
-	dbus_bool_t exit_on_disconnect = TRUE;
+	pid_t pid;
 
 	tdb = pvt;
 	talloc_free(timer);
 	if ((tdb->conn == NULL) ||
 	    !dbus_connection_get_is_connected(tdb->conn)) {
-		/* Close the current connection and open a new one. */
+		/* Close the current connection and exit. */
 		if (tdb->conn != NULL) {
 			dbus_connection_unref(tdb->conn);
+			tdb->conn = NULL;
 		}
-		bus_desc = NULL;
-		switch (tdb->conn_type) {
-		case cm_tdbus_system:
-			cm_log(1, "Attempting to reconnect to system bus.\n");
-			tdb->conn = dbus_bus_get(DBUS_BUS_SYSTEM, NULL);
-			cm_set_conn_ptr(tdb->data, tdb->conn);
-			/* Don't exit if we get disconnected. */
-			exit_on_disconnect = FALSE;
-			bus_desc = "system";
-			break;
-		case cm_tdbus_session:
-			cm_log(1, "Attempting to reconnect to session bus.\n");
-			tdb->conn = dbus_bus_get(DBUS_BUS_SESSION, NULL);
-			cm_set_conn_ptr(tdb->data, tdb->conn);
-			/* Exit if we get disconnected. */
-			exit_on_disconnect = TRUE;
-			bus_desc = "session";
-			break;
-		}
-		if ((tdb->conn != NULL) &&
-		    dbus_connection_get_is_connected(tdb->conn)) {
-			/* We're reconnected; reset our handlers. */
-			cm_log(1, "Reconnected to %s bus.\n", bus_desc);
-			dbus_connection_set_exit_on_disconnect(tdb->conn,
-							       exit_on_disconnect);
-			cm_tdbus_setup_connection(tdb, NULL);
-		} else {
-			/* Try reconnecting again later. */
-			later = tevent_timeval_current_ofs(CM_DBUS_RECONNECT_TIMEOUT, 0),
-			tevent_add_timer(ec, tdb, later,
-					 cm_tdbus_reconnect,
-					 tdb);
-		}
+		pid = getpid();
+		cm_log(0, "Disconnected from dbus, exiting with SIGTERM.\n");
+		kill(pid, SIGTERM);
 	}
 }
 
@@ -493,11 +549,13 @@ cm_tdbus_filter(DBusConnection *conn, DBusMessage *dmessage, void *data)
 {
 	struct tdbus_connection *tdb = data;
 	const char *destination, *unique_name, *path, *interface, *member;
-	/* If we're disconnected, queue a reconnect. */
-	if (!dbus_connection_get_is_connected(conn)) {
+
+	/* If we're disconnected, queue an exit. */
+	if ((tdb->conn_type != cm_tdbus_private) &&
+	    !dbus_connection_get_is_connected(conn)) {
 		tevent_add_timer(talloc_parent(tdb), tdb,
 				 tevent_timeval_current(),
-				 cm_tdbus_reconnect,
+				 cm_tdbus_disconnected,
 				 tdb);
 		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
 	}
@@ -515,7 +573,8 @@ cm_tdbus_filter(DBusConnection *conn, DBusMessage *dmessage, void *data)
 		cm_log(4, "message %p(%s)->%s:%s:%s.%s\n", tdb,
 		       dbus_message_type_to_string(dbus_message_get_type(dmessage)),
 		       destination, path, interface ? interface : "", member);
-		return cm_tdbush_handle_method_call(conn, dmessage, tdb->data);
+		return cm_tdbush_handle_method_call(conn, dmessage,
+						    tdb->conn_type, tdb->data);
 		break;
 	case DBUS_MESSAGE_TYPE_METHOD_RETURN:
 		/* Check that the call or return is directed to us. */
@@ -529,7 +588,9 @@ cm_tdbus_filter(DBusConnection *conn, DBusMessage *dmessage, void *data)
 		       dbus_message_type_to_string(dbus_message_get_type(dmessage)),
 		       (unsigned long) dbus_message_get_reply_serial(dmessage),
 		       (unsigned long) dbus_message_get_serial(dmessage));
-		return cm_tdbush_handle_method_return(conn, dmessage, tdb->data);
+		return cm_tdbush_handle_method_return(conn, dmessage,
+						      tdb->conn_type,
+						      tdb->data);
 		break;
 	default:
 		break;
@@ -538,18 +599,15 @@ cm_tdbus_filter(DBusConnection *conn, DBusMessage *dmessage, void *data)
 }
 
 static int
-cm_tdbus_setup_connection(struct tdbus_connection *tdb, DBusError *error)
+cm_tdbus_setup_conn_loop(struct tdbus_connection *tdb, DBusConnection *conn)
 {
-	DBusError err;
-	const char *bus_desc;
-	int i;
 	/* Set the callback to be called when I/O processing has yielded a
 	 * request that we need to act on. */
-	dbus_connection_set_dispatch_status_function(tdb->conn,
+	dbus_connection_set_dispatch_status_function(conn,
 						     cm_tdbus_dispatch_status,
 						     tdb, NULL);
 	/* Hook up the I/O callbacks so that D-Bus can actually do its thing. */
-	if (!dbus_connection_set_watch_functions(tdb->conn,
+	if (!dbus_connection_set_watch_functions(conn,
 						 &cm_tdbus_watch_add,
 						 &cm_tdbus_watch_remove,
 						 &cm_tdbus_watch_toggle,
@@ -559,7 +617,7 @@ cm_tdbus_setup_connection(struct tdbus_connection *tdb, DBusError *error)
 		return -1;
 	}
 	/* Hook up the (unused?) timer callbacks to be polite. */
-	if (!dbus_connection_set_timeout_functions(tdb->conn,
+	if (!dbus_connection_set_timeout_functions(conn,
 						   cm_tdbus_timeout_add,
 						   cm_tdbus_timeout_remove,
 						   cm_tdbus_timeout_toggle,
@@ -568,57 +626,90 @@ cm_tdbus_setup_connection(struct tdbus_connection *tdb, DBusError *error)
 		cm_log(1, "Unable to add timer callbacks.\n");
 		return -1;
 	}
-	/* Set the filter on messages. */
-	if (!dbus_connection_add_filter(tdb->conn, cm_tdbus_filter,
-					tdb, NULL)) {
+	/* Handle any messages that are already pending. */
+	cm_tdbus_dispatch_status(conn,
+				 dbus_connection_get_dispatch_status(conn),
+				 tdb);
+	return 0;
+}
+
+static int
+cm_tdbus_setup_server_loop(struct tdbus_connection *tdb, DBusServer *server)
+{
+	/* Hook up the I/O callbacks so that D-Bus can actually do its thing. */
+	if (!dbus_server_set_watch_functions(server,
+					     &cm_tdbus_watch_add,
+					     &cm_tdbus_watch_remove,
+					     &cm_tdbus_watch_toggle,
+					     tdb,
+					     &cm_tdbus_watch_cleanup)) {
+		cm_log(1, "Unable to add timer callbacks.\n");
+		return -1;
+	}
+	/* Hook up the (unused?) timer callbacks to be polite. */
+	if (!dbus_server_set_timeout_functions(server,
+					       cm_tdbus_timeout_add,
+					       cm_tdbus_timeout_remove,
+					       cm_tdbus_timeout_toggle,
+					       tdb,
+					       cm_tdbus_timeout_cleanup)) {
+		cm_log(1, "Unable to add timer callbacks.\n");
+		return -1;
+	}
+	return 0;
+}
+
+static int
+cm_tdbus_setup_public_connection(struct tdbus_connection *tdb,
+				 DBusConnection *conn,
+				 const char *bus_desc,
+				 DBusError *error)
+{
+	DBusError err;
+	int ret;
+
+	/* Add the event loop glue. */
+	if (cm_tdbus_setup_conn_loop(tdb, conn) != 0) {
+		cm_log(0, "Error setting up connection to %s bus.\n", bus_desc);
+		return -1;
+	}
+
+	/* Watch for method calls on this connection. */
+	if (!dbus_connection_add_filter(conn, cm_tdbus_filter, tdb, NULL)) {
 		cm_log(1, "Unable to add filter.\n");
 		return -1;
 	}
-	/* Bind to the well-known name we intend to use. */
+
+	/* Request our service name. */
 	memset(&err, 0, sizeof(err));
-	i = dbus_bus_request_name(tdb->conn, CM_DBUS_NAME, 0, &err);
-	if ((i == 0) ||
-	    ((i != DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER) &&
-	     (i != DBUS_REQUEST_NAME_REPLY_ALREADY_OWNER)) ||
+	ret = dbus_bus_request_name(conn, CM_DBUS_NAME, 0, &err);
+	if ((ret == 0) ||
+	    ((ret != DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER) &&
+	     (ret != DBUS_REQUEST_NAME_REPLY_ALREADY_OWNER)) ||
 	    dbus_error_is_set(&err)) {
-		cm_log(-2,
-		       "Unable to set well-known bus name \"%s\": %s(%d).\n",
+		cm_log(0, "Unable to set well-known bus name \"%s\": %s(%d).\n",
 		       CM_DBUS_NAME,
-		       err.message ? err.message : (err.name ? err.name : ""),
-		       i);
+		       err.message ?  err.message : (err.name ? err.name : ""),
+		       ret);
 		if (error != NULL) {
 			dbus_move_error(&err, error);
 		}
 		return -1;
 	}
-	/* Handle any messages that are already pending. */
-	cm_tdbus_dispatch_status(tdb->conn,
-				 dbus_connection_get_dispatch_status(tdb->conn),
-				 tdb);
-	bus_desc = NULL;
-	switch (tdb->conn_type) {
-	case cm_tdbus_system:
-		bus_desc = "system";
-		break;
-	case cm_tdbus_session:
-		bus_desc = "session";
-		break;
-	}
-	cm_log(3, "Connected to %s message bus with name \"%s\", "
-	       "unique name \"%s\".\n",
-	       bus_desc, dbus_bus_get_unique_name(tdb->conn) ?: "(unknown)",
-	       CM_DBUS_NAME);
+	cm_log(3, "Connected to %s message bus with name "
+	       "\"%s\", unique name \"%s\".\n", bus_desc,
+	       dbus_bus_get_unique_name(conn) ?: "(unknown)", CM_DBUS_NAME);
 	return 0;
 }
 
 int
-cm_tdbus_setup(struct tevent_context *ec, enum cm_tdbus_type bus_type,
-	       void *data, DBusError *error)
+cm_tdbus_setup_public(struct tevent_context *ec, enum cm_tdbus_type bus_type,
+		      void *data, DBusError *error)
 {
 	DBusConnection *conn;
+	DBusError err;
 	const char *bus_desc;
 	struct tdbus_connection *tdb;
-	dbus_bool_t exit_on_disconnect;
 
 	/* Build our own context. */
 	tdb = talloc_ptrtype(ec, tdb);
@@ -626,10 +717,10 @@ cm_tdbus_setup(struct tevent_context *ec, enum cm_tdbus_type bus_type,
 		return ENOMEM;
 	}
 	memset(tdb, 0, sizeof(*tdb));
+
 	/* Connect to the right bus. */
 	bus_desc = NULL;
 	conn = NULL;
-	exit_on_disconnect = TRUE;
 	if (error != NULL) {
 		dbus_error_init(error);
 	}
@@ -637,26 +728,162 @@ cm_tdbus_setup(struct tevent_context *ec, enum cm_tdbus_type bus_type,
 	case cm_tdbus_system:
 		conn = dbus_bus_get(DBUS_BUS_SYSTEM, error);
 		cm_set_conn_ptr(data, conn);
-		/* Don't exit if we get disconnected. */
-		exit_on_disconnect = FALSE;
 		bus_desc = "system";
 		break;
 	case cm_tdbus_session:
 		conn = dbus_bus_get(DBUS_BUS_SESSION, error);
 		cm_set_conn_ptr(data, conn);
-		/* Exit if we get disconnected. */
-		exit_on_disconnect = TRUE;
 		bus_desc = "session";
 		break;
+	case cm_tdbus_private:
+		abort();
+		break;
 	}
 	if (conn == NULL) {
-		cm_log(-2, "Error connecting to %s bus.\n", bus_desc);
+		cm_log(0, "Error connecting to %s bus.\n", bus_desc);
 		talloc_free(tdb);
 		return -1;
 	}
-	dbus_connection_set_exit_on_disconnect(conn, exit_on_disconnect);
+	/* Exit on disconnect is handled in cm_tdbus_disconnected(). */
+	dbus_connection_set_exit_on_disconnect(conn, FALSE);
 	tdb->conn = conn;
 	tdb->conn_type = bus_type;
 	tdb->data = data;
-	return cm_tdbus_setup_connection(tdb, error);
+
+	/* Hook up the event loop, register our name, and set up the filter. */
+	memset(&err, 0, sizeof(err));
+	if (cm_tdbus_setup_public_connection(tdb, conn, bus_desc, &err) != 0) {
+		talloc_free(tdb);
+		return -1;
+	}
+
+	return 0;
+}
+
+static void
+cm_tdbus_new_private_client(DBusServer *server, DBusConnection *new_conn,
+			    void *data)
+{
+	struct tdbus_connection *tdb = data;
+	int sd;
+
+	if (dbus_connection_get_socket(new_conn, &sd)) {
+		cm_log(4, "New client on FD %d.\n", sd);
+	} else {
+		cm_log(4, "New client on unknown socket.\n");
+	}
+	if (cm_tdbus_setup_conn_loop(tdb, new_conn) == 0) {
+		/* Watch for method calls on this connection. */
+		if (!dbus_connection_add_filter(new_conn, cm_tdbus_filter,
+						tdb, NULL)) {
+			cm_log(1, "Unable to add filter, dropping.\n");
+			return;
+		}
+		dbus_connection_ref(new_conn);
+		cm_log(3, "Accepted private connection.\n");
+	} else {
+		cm_log(0, "Error setting up for client, dropping.\n");
+	}
+}
+
+static void
+cm_tdbus_lost_private_client(void *data)
+{
+	cm_log(3, "Lost private connection.\n");
+}
+
+#ifndef HAVE_OPENSSL
+#ifdef HAVE_GMP
+static void
+fill_uuid(unsigned char *uuid, size_t length)
+{
+	gmp_randstate_t state;
+	unsigned int i;
+
+	gmp_randinit_default(state);
+	for (i = 0; i < length; i++) {
+		uuid[i] = gmp_urandomb_ui(state, 8);
+	}
+}
+#endif
+#endif
+
+int
+cm_tdbus_setup_private(struct tevent_context *ec, void *data,
+		       const char *path, char **address, DBusError *error)
+{
+	struct tdbus_connection *tdb;
+	unsigned char uuid[16];
+	char *addr;
+
+	*address = NULL;
+
+	/* Build our own context. */
+	tdb = talloc_ptrtype(ec, tdb);
+	if (tdb == NULL) {
+		return ENOMEM;
+	}
+	memset(tdb, 0, sizeof(*tdb));
+
+	/* Start up the listener. */
+	if (error != NULL) {
+		dbus_error_init(error);
+	}
+	if (path != NULL) {
+		if (path[0] == '/') {
+			addr = talloc_asprintf(ec, "unix:path=%s", path);
+		} else {
+			addr = talloc_asprintf(ec, "unix:%s", path);
+		}
+	} else {
+#ifdef HAVE_UUID
+		if (cm_submit_uuid_new(uuid) == 0) {
+			/* we're good */
+		} else
+#endif
+#ifdef HAVE_OPENSSL
+		if (!RAND_bytes(uuid, sizeof(uuid))) {
+			/* Try again sometime later. */
+			cm_log(1, "Error generating UUID.\n");
+			talloc_free(tdb);
+			return -1;
+		}
+#else
+#ifdef HAVE_GMP
+		fill_uuid(uuid, sizeof(uuid));
+#endif
+#endif
+		addr = talloc_asprintf(ec, "unix:abstract=%s/listen-"
+				       "%02x%02x%02x%02x%02x%02x%02x%02x"
+				       "%02x%02x%02x%02x%02x%02x%02x%02x",
+				       CM_TMPDIR,
+				       uuid[0], uuid[1], uuid[2], uuid[3],
+				       uuid[4], uuid[5], uuid[6], uuid[7],
+				       uuid[8], uuid[9], uuid[10], uuid[11],
+				       uuid[12], uuid[13], uuid[14], uuid[15]);
+	}
+	tdb->server = dbus_server_listen(addr, error);
+	if (dbus_error_is_set(error)) {
+		cm_log(0, "Error setting up D-Bus server.\n");
+		talloc_free(tdb);
+		return -1;
+	}
+	tdb->conn_type = cm_tdbus_private;
+	tdb->data = data;
+
+	/* Add the event loop glue. */
+	if (cm_tdbus_setup_server_loop(tdb, tdb->server) != 0) {
+		cm_log(0, "Error setting up private listener.\n");
+		talloc_free(tdb);
+		return -1;
+	}
+
+	/* Provide the callback to use when we get a new client connection. */
+	dbus_server_set_new_connection_function(tdb->server,
+						cm_tdbus_new_private_client,
+						tdb,
+						cm_tdbus_lost_private_client);
+
+	*address = dbus_server_get_address(tdb->server);
+	return 0;
 }
